(* ========================================================================= *)
(*                                                                           *)
(*                           tactics_1.ml                              *)
(*                                                                           *)
(*                                                                           *)
(*                                                                           *)
(* Last update: October 24, 2016                                            *)
(*                                                                           *)
(* ========================================================================= *)


(*-------------REQUIRED LIBRARIES ---------------------*)


needs "tensor_product_projection.ml";;



(*-----------------------------------------------------*)
(*----------------Formalization------------------------*)
(*-----------------------------------------------------*)
(*--------------------**********************------------------*) 

let spec_thm_tac thm = 
(CONV_RULE (RAND_CONV (RAND_CONV 
 (ONCE_REWRITE_CONV[CFUN_ARITH ` ((X:sm->(real->complex)) (x:sm)) = 
 Cx(&1) % ((X:sm->(real->complex)) (x:sm))`]))) (SPEC_ALL thm));;
 
(****************************************************************************************)
 
 let condition_thm n1 n3 = 
    let rec make_concl n i = if (i<n) then 
        "if i = " :: (string_of_int i) :: " then x" :: 
        (string_of_int i) :: " else " :: (make_concl n (i+1)) 
        else  "x" :: (string_of_int i) :: " else " :: ["g i):A"] in
    let make_cond n1 n2 = "(if i <= " :: (string_of_int n1) :: 
        " /\ 1 <= i then " :: (make_concl n2 1) in
    let main_make_cond n1 n2 = parse_term (String.concat "" (make_cond n1 n2)) in 
    ARITH_RULE  (mk_eq ((main_make_cond n1 n3), (main_make_cond n1 n1)));;
        
(****************************************************************************************)
        
let condition_thm0 n1 n3 = 
    let rec make_concl n i = if (i<n) then 
        "if i = " :: (string_of_int i) :: " then x" :: 
        (string_of_int i) :: " else " :: (make_concl n (i+1)) 
        else  "x" :: (string_of_int i) :: " else " :: ["g i):A"] in
    let make_cond0 n1 n2 = "(if i <= " :: (string_of_int n1) :: 
        " /\ 1 <= i then " :: (make_concl n2 0) in
    let make_cond n1 n2 = "(if i <= " :: (string_of_int n1) :: 
        " /\ 1 <= i then " :: (make_concl n2 1) in
    let main_make_cond n1 n2 = parse_term (String.concat "" (make_cond n1 n2)) in 
    let main_make_cond0 n1 n2 = parse_term (String.concat "" (make_cond0 n1 n2)) in 
    ARITH_RULE  (mk_eq ((main_make_cond0 n1 n3), (main_make_cond n1 n1)));;

    
(****************************************************************************************)  
         
let rec get_suc  =  function 
|0 -> "0"
|n when n > 0  -> String.concat "" ["SUC" ;"(";(get_suc (n-1));")"];; 

let rec generat  = function
|0 -> []
|n when n > 0 ->  (ARITH_RULE(parse_term 
 (String.concat "" [get_suc n;"=";string_of_int n])))::(generat (n-1))  ;;

let lem1 =  GEN_ALL(ARITH_RULE ` 
  (if i <= n1 /\ n2 <= i then if i = n2 then a else c else e) =
 (if i <= n1 /\ n2 <= i /\ i = n2 then a else if i <= n1 /\ 
  SUC n2 <= i then c else e)`);;
 
 let lem2 =  ARITH_RULE `!(i:num) (n2:num) a c e. 
  (if i <= n2 /\ n2 <= i then if i = n2 then a else c else e) =
         (if i = n2 then a else e)`;;
         
let lem3 = ARITH_RULE`!(i:num) n2 a b.
   (if i <= n2 /\ n2 <= i then a else b) = (if i = n2 then a else b)`;;

 let rec generat2 n  = function
|[] -> []
|h::t  ->  (ARITH_RULE(parse_term 
(String.concat "" [get_suc (n-h);"<=";get_suc n])))::(generat2 (n-h) t)  ;;

 let rec generat3 n  = function
|0 -> [LE_0]
|m when 0 < m   ->  (ARITH_RULE(parse_term 
(String.concat "" [get_suc m;"<=";get_suc n])))::(generat3 n (m-1))  ;;

let rec generat4 n = function
|[] -> []
|h::t -> (generat3 (n-h) (n-h)) @(generat4 (n-h) t);;

let lem7 =  (ARITH_RULE `!(i:num) (n2:num) a b. n4 <= n3 ==> 
(if (i <= n3 /\ n2 <= i) /\ i <= n4 then a  else b) =
 (if i <= n4 /\ n2 <= i then a else b)`);;

let lem5 =  GEN_ALL(MESON[]` 
(if p then if q then a else c else e) = 
(if (p /\ q) then a else if (p /\ ~q)  then c else e)`);;

let lem6 =  (ARITH_RULE `!(i:num) (n2:num) a b. 
 (if (i <= n3 /\  SUC 0 <= i) /\ ~(i <=  (SUC n2)) then a  else b) =
 (if i <= n3 /\  (SUC (SUC n2)) <= i then a else b)`);;


 let lem8 =  ARITH_RULE `!i n1 n2:num. n2 <= n1 ==> 
  (i <= n1 /\n2 <=  i /\ i = n2 <=> i = n2)`;;

 let lem10 = ARITH_RULE `!i n1 n2 n3:num. (if i <= n1 /\n2 <=  i /\
 i = n2 then a else if i <= n1 /\ SUC n2 <= i then b else c )
  <=> (if  i <= n1 /\ SUC n2 <=  i  then b else if  i <= n1 /\ 
  n2 <=  i /\ i = n2 then a else c)`;;
 
         
let condition_recompose m l  = 
    let rec make_concl1 n i = if (i<n) then 
        "if i = " :: (get_suc i) :: " then x" :: (string_of_int i) :: 
        " else " :: (make_concl1 n (i+1)) 
        else  "x" :: (string_of_int i) :: " else " :: ["g i):A"] in
    let make_cond1 n1 n2 = "(if i <= " :: (get_suc n1) :: 
        " /\ SUC 0 <= i then " :: (make_concl1 n2 1) in
    let main_make_cond n1 n2 = parse_term (String.concat "" (make_cond1 n1 n2)) in
    let rec make_concl k n i = if (i<n) then 
        "if i = " :: (get_suc (k+i)) :: " then x" :: 
        (string_of_int (k+i)) :: " else " :: (make_concl k n (i+1)) 
        else  "x" :: (string_of_int (k+i)) :: [""] in
    let rec main_make_concl l i k = 
        let len = List.length l in
            if (i<len) then
                (make_concl k (List.nth l i) 1) @ [" else " ] @ 
                (main_make_concl l (i+1) (k+(List.nth l i)))
            else [] in
        let rec make_fst_cond l i k = 
            let len = List.length l in
            if (i<(len-1)) then
            (make_fst_cond l (i+1) (k+(List.nth l i))) @ 
            ("if i <= " :: (get_suc (k+(List.nth l i))) :: " then ":: []) 
            else [""] in
    REWRITE_RULE (generat m) (prove( (mk_eq ((parse_term (String.concat ""  ("(if i <= " :: 
    (get_suc m) :: " /\ SUC 0 <= i then " :: 
    (make_fst_cond l 0 0) @ (main_make_concl l 0 0) @ [" g i):A"]))), (main_make_cond m m))),
    let lgenerat2 = generat2 m (rev l)  in
 REPEAT (CHANGED_TAC (ONCE_REWRITE_TAC[lem1] THEN ONCE_REWRITE_TAC[lem2])) 
 THEN REWRITE_TAC[lem3]
 THEN REPEAT(CHANGED_TAC (ONCE_REWRITE_TAC [lem5] THEN 
  REWRITE_TAC[lem6] THEN SIMP_TAC([lem7]@lgenerat2)
 THEN REPEAT (CHANGED_TAC (ONCE_REWRITE_TAC[lem1] THEN ONCE_REWRITE_TAC[lem2])) 
 THEN REWRITE_TAC[lem3])) THEN
 SIMP_TAC([lem8]@(generat4 m (rev l))@(generat3 m (m-1)))));;
 
    
    
(****************************************************************************************)
            
let rew_condition_tac m l i = 
    let rec rew_condition m l i = 
        let len = List.length l in
        if (i < len)  then
            if (i = 0) then 
             (condition_thm (List.nth l i) m) :: (rew_condition (m-(List.nth l i)) l (i+1))
            else
             (condition_thm0 (List.nth l i) m) :: (rew_condition (m-(List.nth l i)) l (i+1))
        else
            [] in
        SIMP_TAC (rew_condition m l i);;
        
(****************************************************************************************)
        
let one_less n = 
    let thm_integer n =  
        parse_term (String.concat "" ("(1 <= " :: (string_of_int n) :: ["):bool"])) in
     map ARITH_RULE (map thm_integer (1--n));;

(****************************************************************************************)   
    
let rewrite_decompose_tac m l k i =
    let rec rewrite_decompose m l k i =  
       let len = List.length l in
       let thm_integer n m k = "((" :: (string_of_int m) :: " <= dimindex (:N) ==> (i <= " ::
       (string_of_int n) :: " ==> i + " :: (string_of_int k) :: " <= dimindex (:N))) /\ (1 <= i + " :: 
       (string_of_int k) :: [")):bool"] in 
        if (i < len)  then
            if (i = 0) then 
                (rewrite_decompose m l (k+(List.nth l i)) (i+1))
            else
            (parse_term (String.concat "" (thm_integer (List.nth l i) m k))) :: 
            (rewrite_decompose m l (k+(List.nth l i)) (i+1))
        else [] in  
    (ASM_SIMP_TAC ([ARITH_RULE `(((i:num) + j = k) <=> (if (j <= k) then i = k - j else F))`;
    LAMBDA_BETA] @ map ARITH_RULE (rewrite_decompose m l k i))) THEN CONV_TAC NUM_REDUCE_CONV;;
    
(****************************************************************************************)
     
let rewrite_recompose_tac m l k i =
    let rec rewrite_recompose m l k i =  
       let len = List.length l in
       let thm_integer n k  = "((("::(string_of_int n)::" <= dimindex (:N)) ==> (i <= "::
       (string_of_int n)::" ==> ((i - ":: (string_of_int k)::") <= dimindex (:N)))) /\ (~(i <= ":: 
       (string_of_int k):: ")  ==> (1 <= (i - ":: (string_of_int k)::[" )))):bool "] in 
        if (i < len)  then
           (parse_term (String.concat "" (thm_integer m k))) :: 
           (rewrite_recompose m l (k+(List.nth l i)) (i+1))
        else [] in
    ONCE_SIMP_TAC [(ISPECL [mk_numeral (Int m)] tensor_1mlem1d)] THEN       
    (ASM_SIMP_TAC ([ARITH_RULE `~(i <= j) ==> (((i:num) - j = k) <=> (i = k + j))`;
    LAMBDA_BETA] @ (map (REWRITE_RULE [SUB_0]) 
    (map ARITH_RULE (rewrite_recompose m l k i))))) THEN 
    CONV_TAC NUM_REDUCE_CONV ;; 
    
(****************************************************************************************)  


let read_file filename = 
let lines = ref [] in
let chan = Pervasives.open_in filename in
try
  while true; do
    lines := input_line chan :: !lines
  done; !lines
with End_of_file ->
  close_in chan;
  List.rev !lines ;;
  
(****************************************************************************************)   
  
let rec extract_port str0 str i j = 
    let len = String.length str in
    if (i < len) then 
      if ((((Char.code (String.get str i)) <= 122) && ((Char.code (String.get str i)) >= 97)) or 
        (((Char.code (String.get str i)) <= 90) && ((Char.code (String.get str i)) >= 65)) or 
        (((Char.code (String.get str i)) <= 57) && ((Char.code (String.get str i)) >= 48))) then
        
            (extract_port str0 str (i+1) (j+1))
            
      else if ((String.sub str i 1) = ",") then
         if (((String.sub str (i+1) 1) = "0") or ((String.sub str (i+1) 1) = "1")) then
         
            (extract_port (str0 @ [(String.sub str (i-j) j)]) str (i+2) 0)  
            
         else 
             (extract_port str0 str (i+1) j)
      else
         (extract_port str0 str (i+1) j)
    else
            str0;;

(****************************************************************************************)

let rec gate_column_sub str0 str new_circuits i j k   = 
      let element = fst (List.nth new_circuits k) in
      let element_size = snd (List.nth new_circuits k)  in
      let len  = String.length element in
      let size = List.length new_circuits in
      if (k < size) then
       if ((String.sub str 0 len) = element) then 
        (String.sub str 0 len)::string_of_int(element_size)::str0 @ [(String.sub str (i-j) j)]
       else 
        gate_column_sub str0 str new_circuits i j (k+1)
      else [];;
         
  
let rec gate_column str0 str new_circuits i j =
    let len = String.length str in
    if (i <= len)  then
      if ((String.sub str i 1) = ")")  then
        if ((String.sub str 0 9) = "FLIP_GATE") then
               (String.sub str 0 9)::"1"::str0 @ [(String.sub str (i-j) j)]
        else if ((String.sub str 0 13) = "HADAMARD_GATE") then
                (String.sub str 0 13)::"1"::str0 @ [(String.sub str (i-j) j)]
        else if (((String.sub str 0 7) = "CZ_GATE")) then
                 (String.sub str 0 7)::"2"::str0 @ [(String.sub str (i-j) j)]
        else if (((String.sub str 0 10) = "CNOT1_GATE") or ((String.sub str 0 10) = "CNOT2_GATE")) then 
                (String.sub str 0 10)::"2"::str0 @ [(String.sub str (i-j) j)]
        else if ((String.sub str 0 9) = "SWAP_GATE") then
                (String.sub str 0 9)::"2"::str0 @ [(String.sub str (i-j) j)]
        else if ((String.sub str 0 7) = "TS_GATE") then
                 (String.sub str 0 7)::"3"::str0 @ [(String.sub str (i-j) j)]
        else if (((String.sub str 0 13) = "TOFFOLI1_GATE") or ((String.sub str 0 13) = "TOFFOLI3_GATE") or
                ((String.sub str 0 13) = "FREDKIN1_GATE") or ((String.sub str 0 13) = "FREDKIN3_GATE")) then
                 (String.sub str 0 13)::"3"::str0 @ [(String.sub str (i-j) j)]
        else
            gate_column_sub str0 str new_circuits i j 0
      else
        if ((((Char.code (String.get str i)) <= 122) && ((Char.code (String.get str i)) >= 97)) or 
         (((Char.code (String.get str i)) <= 90) && ((Char.code (String.get str i)) >= 65)) or 
         (((Char.code (String.get str i)) <= 57) && ((Char.code (String.get str i)) >= 48))) then
            gate_column str0 str new_circuits (i+1) (j+1)
            else if ((String.sub str i 1) = ",") then
                gate_column (str0 @ [(String.sub str (i-j) j)]) str new_circuits (i+1) 0
            else
                gate_column str0 str new_circuits (i+1) 0
        else [];;
        
(****************************************************************************************)      
        
let rec gate_matrix file str new_circuits i  =
    let gates = read_file file in
    let len = List.length gates in
        if ( i < len) then
            gate_matrix file (str @ [(gate_column [] (List.nth gates i) new_circuits 0 0)]) new_circuits  (i+1)
        else
            str     ;;

(****************************************************************************************)          
            
let rec list_replace L n a= 
    match L with
    | [] -> []
    | (hd::l) -> if n = 0 then
                    a::l
                else
                    hd :: (list_replace l (n-1) a);;
                    
                    
                    
let rec connected_gates gate ports j i k =  
            if (j < (k+2)) then
             if (j = 1) then
              if (int_of_string (List.nth gate j) = k) then
               (connected_gates gate ports (j+1) i k) else  false
              else 
               if ( (List.nth gate j) = (List.nth ports i)) then
                (connected_gates gate ports (j+1) (i+1) k) else false
             else  true;;
             
    let rec list_replace_list gate ports j i k =
                if (j < k) then
                   list_replace_list gate (list_replace ports i (List.nth gate (2+k+j))) (j+1) (i+1) k
                else ports;;
                
let rec sub_procedure_1 gates ports  i j k ks = 
      let len_gates = List.length gates in
       if j < len_gates then
         if (k < (ks+1)) then
            if (connected_gates (List.nth gates j) ports 1 i k) then 
                [((string_of_int i) :: (List.nth gates j));(list_replace_list (List.nth gates j) ports 0 i k)]
            else 
                sub_procedure_1 gates ports  i j (k+1) ks
        else sub_procedure_1 gates ports  i (j+1) 1 ks
        else [];;
    
let rec sub_procedure_2 gates ports i ks fin =
        let len_gates = List.length gates in
        let len_port = List.length ports in
          if (ks > 0) then  
            if i <= (len_port - ks) then
               let sub2_result = sub_procedure_1 gates ports  i 0 1 ks in 
                 if sub2_result = [] then
                   sub_procedure_2 gates ports (i+1) ks fin
                 else 
   [(List.nth sub2_result 0)] @ ((sub_procedure_2 gates (List.nth sub2_result 1) (i+(int_of_string (List.nth (List.nth sub2_result 0) 2))) ks (fin+1)) )
            else 
                sub_procedure_2 gates ports i (ks-1) fin
           else [];;
           
let rec sub_procedure_3 gates ports i ks fin =
        let len_gates = List.length gates in
        let len_port = List.length ports in
          if (ks > 0) then  
            if i <= (len_port - ks) then
               let sub2_result = sub_procedure_1 gates ports  i 0 1 ks in 
                 if sub2_result = [] then
                   sub_procedure_3 gates ports (i+1) ks fin
                 else 
                  (sub_procedure_3 gates (List.nth sub2_result 1) (i+(int_of_string (List.nth (List.nth sub2_result 0) 2))) ks (fin+1))
            else 
                sub_procedure_3 gates ports i (ks-1) fin
           else fin;;

let rec sub_procedure_4 gates ports i ks fin =
        let len_gates = List.length gates in
        let len_port = List.length ports in
          if (ks > 0) then  
            if i <= (len_port - ks) then
               let sub2_result = sub_procedure_1 gates ports  i 0 1 ks in 
                 if sub2_result = [] then
                   sub_procedure_4 gates ports (i+1) ks fin
                 else 
                  (sub_procedure_4 gates (List.nth sub2_result 1) (i+(int_of_string (List.nth (List.nth sub2_result 0) 2))) ks (fin+1))
            else 
                sub_procedure_4 gates ports i (ks-1) fin
           else ports;;
           
let rec matrix_procedure str0 gates ports ks = 
    let fin = sub_procedure_3 gates ports 0 ks 0 in 
         if (fin > 0) then 
         matrix_procedure (str0 @ [sub_procedure_2 gates ports 0 ks 0]) gates (sub_procedure_4 gates ports 0 ks 0) ks
        else str0;;
                    
(****************************************************************************************)
let rec list_sum list = match list with
    | [] -> 0
    | head::tail -> head + list_sum tail;;
    

(****************************************************************************************)
let main_tab_gates l m =
    let rec tab_gates l i =
       let len = List.length l in
       let mem_list k = (List.nth (List.nth l i) k) in
       let mem_list1 k = (List.nth (List.nth l (i-1)) k) in
       if (i < len)  then 
            if  (i = 0)  then
                if ((int_of_string (mem_list 0)) = 0) then
                   (int_of_string (mem_list 2)) :: (tab_gates l (i+1))
                else 
                   (int_of_string (mem_list 0)) :: (int_of_string (mem_list 2)) :: (tab_gates l (i+1))
            else
                if ( ((int_of_string (mem_list1 0)) + (int_of_string (mem_list1 2))) = 
                     (int_of_string (mem_list 0))) then
                    (int_of_string (mem_list 2)) :: (tab_gates l (i+1))
                else
                    ((int_of_string (mem_list 0)) - (int_of_string (mem_list1 2)) - 
                    (int_of_string (mem_list1 0))) :: (int_of_string (mem_list 2)) ::
                    (tab_gates l (i+1))
       else [] in
    let k = tab_gates l 0 in
    if ((list_sum k) < m) then
       k @ [m-(list_sum k)]
    else    k;;

(****************************************************************************************)
let main_comp_inputs l m =
    let rec rewr_para_dev n k = 
            if (k < n)  then "(":: (rewr_para_dev n (k+1))
            else [] in
    let rec comp_inputs l m i =
      let len = List.length l in
      if (i < len)  then 
        if (i=0) then
         ((string_of_int m) :: "  = " :: (rewr_para_dev len 1)) @ 
         ((string_of_int (List.nth l i)):: (comp_inputs l m (i+1)))
        else
          " + " :: (string_of_int (List.nth l i)):: " ) " :: (comp_inputs l m (i+1))
      else [] in
      ARITH_RULE (parse_term (String.concat "" (comp_inputs l m 0)));;



(****************************************************************************************)
    
   let rec rewr_dev m i =    
        let rec rewr_para_dev n k = 
          if (k < n) then ("(" :: (rewr_para_dev n (k+1)))
          else  [] in
        let rec rewr_conc_orig n k = 
          if (k < n) then if(k=0) then
          (rewr_para_dev n 1) @ (" ((f"::(string_of_int k)::") x"::(string_of_int k)::
          ")" :: (rewr_conc_orig n (k+1)))
           else " * ((f"::(string_of_int k)::") x"::(string_of_int k)::")) " :: 
           (rewr_conc_orig n (k+1))
          else  [] in
        let rec rewr_sing_orig n k = 
          if (k < n) then if(k=0) then
          " ((f"::(string_of_int k)::") x"::(string_of_int k)::")" :: (rewr_sing_orig n (k+1))
           else " * ((f"::(string_of_int k)::") x"::(string_of_int k)::") " :: 
           (rewr_sing_orig n (k+1))
          else  [] in
        let rec rewr_conc_dev n k = 
          if (k < n) then if(k=0) then
           [" a0 * ("] @ (" ((f"::(string_of_int k)::") x"::
           (string_of_int k)::")" :: (rewr_conc_dev n (k+1)))
          else " * ((f"::(string_of_int k)::") x"::(string_of_int k)::") " :: 
          (rewr_conc_dev n (k+1))
          else  [")"] in
        let rec rewr_sing_dev n k l = 
          if (k < n) then if(k=0) then if (k=l) then 
            "  ((a0 % f0) x0) "::(rewr_sing_dev n (k+1) l)
           else "  ((f0) x0) "::(rewr_sing_dev n (k+1) l)
          else if (k=l) then 
           "* ((a0 % f"::(string_of_int k)::") x"::(string_of_int k)::") "::
           (rewr_sing_dev n (k+1) l)
           else "* ((f"::(string_of_int k)::") x"::(string_of_int k)::") "::
           (rewr_sing_dev n (k+1) l)
          else [] in
        if (i < m) then
        (CFUN_ARITH (parse_term (String.concat "" ( ["! (f0:A^N->complex) a0. ("] @ (rewr_sing_dev m 0 i) @ [" = "] @ 
        (rewr_conc_dev m 0) @ [")"])) )) :: (rewr_dev m (i+1))
        else [(CFUN_ARITH (parse_term (String.concat "" ( ["! (f0:A^N->complex). ("] @ (rewr_sing_orig m 0) @ 
        [" = "] @ (rewr_conc_orig m 0)  @ [")"]))))];;
        
(****************************************************************************************)
     
let rewrite_l l =      
   let spec_k k = (ISPECL [mk_numeral (Int k)] tensor_1mlem1d) in 
        map spec_k l;;

(****************************************************************************************)
           
let integer_equiv n = 
    let thm_integer = ARITH_RULE `!n. n <= dimindex (:N) <=> 
    (n <= dimindex (:N) /\ (n-1) <= dimindex (:N))` in
    let spec_new thm x =  SPEC  (mk_numeral (Int x))  thm in 
    let once_acm_rewrite_new thm  = ONCE_ASM_REWRITE_TAC [thm] in 
    MAP_EVERY once_acm_rewrite_new (map (CONV_RULE NUM_REDUCE_CONV) 
    (rev (map (spec_new  thm_integer) (2--n))));;
    
(****************************************************************************************)
        
let CFUN_ADD_RDISTRIB_NEW = CFUN_ARITH 
`!(a:complex) (b:complex) (x:cfun) (x1:cfun) (x2:cfun). 
(a + b) % x = a % x + b % x /\  
(a + b) % x + x1 = a % x + b % x + x1 /\ 
x2 + (a + b) % x + x1= x2+ a % (x:cfun) + b % x + x1`;;



let thm1 =  MESON[REAL_OF_NUM_MUL;REAL_MUL_ASSOC;REAL_MUL_SYM]
`&m * &n * x = &(m * n) * x /\ &m * y * &n * x = &(m * n) * x * y /\ &m * &n = &(m * n) /\
z * &n  = &n * z /\ &n  * z * &m  = &(m * n) * z /\
z * &n * z2  = &n * z * z2   `;;

let thm2 =  REAL_FIELD `(y1:real) / x1 * y2 / x2 * z = (y1 * y2) / (x1 * x2) * z /\
y1 / x1 * z * y2 / x2  = (y1 * y2) / (x1 * x2) * z /\ 
z * (y1 / x1)  = (y1) / (x1) * z /\ 
z * (y1 / x1) * z2  = (y1) / (x1) * z * z2 /\ 
y1 / x1 * z1 * y2 / x2 * z2 = (y1 * y2) / (x1 * x2) * z1 * z2 /\  
y1 / x1 * y2 / x2 = (y1 * y2) / (x1 * x2)`;;


(****************************************************************************************)


    